#!/usr/bin/env python3

import argparse
import json

import singer

from tap_exacttarget.state import load_state, save_state

from tap_exacttarget.client import get_auth_stub

from tap_exacttarget.endpoints.campaigns \
    import CampaignDataAccessObject
from tap_exacttarget.endpoints.content_areas \
    import ContentAreaDataAccessObject
from tap_exacttarget.endpoints.data_extensions \
    import DataExtensionDataAccessObject
from tap_exacttarget.endpoints.emails import EmailDataAccessObject
from tap_exacttarget.endpoints.events import EventDataAccessObject
from tap_exacttarget.endpoints.folders import FolderDataAccessObject
from tap_exacttarget.endpoints.lists import ListDataAccessObject
from tap_exacttarget.endpoints.list_sends import ListSendDataAccessObject
from tap_exacttarget.endpoints.list_subscribers \
    import ListSubscriberDataAccessObject
from tap_exacttarget.endpoints.sends import SendDataAccessObject
from tap_exacttarget.endpoints.subscribers import SubscriberDataAccessObject


LOGGER = singer.get_logger()  # noqa


def validate_config(config):
    required_keys = ['client_id', 'client_secret']
    missing_keys = []
    null_keys = []
    has_errors = False

    for required_key in required_keys:
        if required_key not in config:
            missing_keys.append(required_key)

        elif config.get(required_key) is None:
            null_keys.append(required_key)

    if missing_keys:
        LOGGER.fatal("Config is missing keys: {}"
                     .format(", ".join(missing_keys)))
        has_errors = True

    if null_keys:
        LOGGER.fatal("Config has null keys: {}"
                     .format(", ".join(null_keys)))
        has_errors = True

    if has_errors:
        raise RuntimeError


def load_catalog(filename):
    catalog = {}

    try:
        with open(filename) as handle:
            catalog = json.load(handle)
    except Exception:
        LOGGER.fatal("Failed to decode catalog file. Is it valid json?")
        raise RuntimeError

    return catalog


def load_config(filename):
    config = {}

    try:
        with open(filename) as handle:
            config = json.load(handle)
    except Exception:
        LOGGER.fatal("Failed to decode config file. Is it valid json?")
        raise RuntimeError

    validate_config(config)

    return config


AVAILABLE_STREAM_ACCESSORS = [
    CampaignDataAccessObject,
    ContentAreaDataAccessObject,
    DataExtensionDataAccessObject,
    EmailDataAccessObject,
    EventDataAccessObject,
    FolderDataAccessObject,
    ListDataAccessObject,
    ListSendDataAccessObject,
    ListSubscriberDataAccessObject,
    SendDataAccessObject,
    SubscriberDataAccessObject,
]


def do_discover(args):
    LOGGER.info("Starting discovery.")

    config = load_config(args.config)
    state = load_state(args.state)

    auth_stub = get_auth_stub(config)

    catalog = []

    for available_stream_accessor in AVAILABLE_STREAM_ACCESSORS:
        stream_accessor = available_stream_accessor(
            config, state, auth_stub, None)

        catalog += stream_accessor.generate_catalog()

    print(json.dumps({'streams': catalog}))


def _is_selected(catalog_entry):
    default = catalog_entry.get('selected-by-default', False)

    return ((catalog_entry.get('inclusion') == 'automatic') or
            (catalog_entry.get('inclusion') == 'available' and
             catalog_entry.get('selected', default) is True))


def do_sync(args):
    LOGGER.info("Starting sync.")

    config = load_config(args.config)
    state = load_state(args.state)
    catalog = load_catalog(args.properties)

    auth_stub = get_auth_stub(config)

    stream_accessors = []

    subscriber_selected = False
    subscriber_catalog = None
    list_subscriber_selected = False

    for stream_catalog in catalog.get('streams'):
        stream_accessor = None

        if not _is_selected(stream_catalog.get('schema', {})):
            LOGGER.info("'{}' is not marked selected, skipping."
                        .format(stream_catalog.get('stream')))
            continue

        if SubscriberDataAccessObject.matches_catalog(stream_catalog):
            subscriber_selected = True
            subscriber_catalog = stream_catalog
            LOGGER.info("'subscriber' selected, will replicate via "
                        "'list_subscriber'")
            continue

        if ListSubscriberDataAccessObject.matches_catalog(stream_catalog):
            list_subscriber_selected = True

        for available_stream_accessor in AVAILABLE_STREAM_ACCESSORS:
            if available_stream_accessor.matches_catalog(stream_catalog):
                stream_accessors.append(available_stream_accessor(
                    config, state, auth_stub, stream_catalog))

                break

    if subscriber_selected and not list_subscriber_selected:
        LOGGER.fatal('Cannot replicate `subscriber` without '
                     '`list_subscriber`. Please select `list_subscriber` '
                     'and try again.')
        exit(1)

    for stream_accessor in stream_accessors:
        if isinstance(stream_accessor, ListSubscriberDataAccessObject) and \
           subscriber_selected:
            stream_accessor.replicate_subscriber = True
            stream_accessor.subscriber_catalog = subscriber_catalog

        try:
            stream_accessor.state = state
            stream_accessor.sync()
            state = stream_accessor.state

        except Exception:
            LOGGER.error('Failed to sync endpoint, moving on!')

    save_state(state)


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-c', '--config', help='Config file', required=True)
    parser.add_argument(
        '-s', '--state', help='State file')
    parser.add_argument(
        '-p', '--properties', help='Catalog file with fields selected')

    parser.add_argument(
        '-d', '--discover',
        help='Build a catalog from the underlying schema',
        action='store_true')
    parser.add_argument(
        '-S', '--select-all',
        help=('When "--discover" is set, this flag selects all fields for '
              'replication in the generated catalog'),
        action='store_true')

    args = parser.parse_args()

    try:
        if args.discover:
            do_discover(args)
        elif args.properties:
            do_sync(args)
        else:
            LOGGER.info("No properties were selected")
    except RuntimeError as exception:
        LOGGER.error(str(exception))
        LOGGER.fatal("Run failed.")
        exit(1)


if __name__ == '__main__':
    main()



from suds.mx.appender import Appender,Content


class _PropertyAppender(Appender):
    """
    A L{Property} appender.

    Patched for FuelSDK
    """
        
    def append(self, parent, content):
        p = content.value
        child = self.node(content)
        child_value = p.get()
        if(child_value is None):
            pass
        else:
            child.setText(child_value)
            parent.append(child)
            for item in list(p.items()):
                cont = Content(tag=item[0], value=item[1])
                Appender.append(self, child, cont)


def _bodycontent(self, method, args, kwargs):
    """
    # The I{wrapped} vs I{bare} style is detected in 2 ways.
    # If there is 2+ parts in the message then it is I{bare}.
    # If there is only (1) part and that part resolves to a builtin then
    # it is I{bare}.  Otherwise, it is I{wrapped}.
    #

    Patched for FuelSDK
    """
    if not len(method.soap.input.body.parts):
        return ()
    wrapped = method.soap.input.body.wrapped
    if wrapped:
        pts = self.bodypart_types(method)
        root = self.document(pts[0])
    else:
        root = []
    n = 0
    for pd in self.param_defs(method):
        if n < len(args):
            value = args[n]
        else:
            value = kwargs.get(pd[0])
        n += 1
        if value is None:
            continue
        p = self.mkparam(method, pd, value)
        if p is None:
            continue
        if not wrapped:
            ns = pd[1].namespace('ns0')
            p.setPrefix(ns[0], ns[1])
        root.append(p)
    return root



import requests
import json
import copy

    
########
##
##  Parent class used to determine what status we are in depending on web service call results
##
########
class ET_Constructor(object):
    results = []
    code = None
    status = False
    message = None
    more_results = False                
    request_id = None
    
    def __init__(self, response = None, rest = False):
        
        if response is not None:    #if a response was returned from the web service call
            if rest:    # result is from a REST web service call...
                self.code = response.status_code            
                if self.code == 200:
                    self.status = True
                else:
                    self.status = False 
                        
                try:
                    self.results = response.json()
                except:
                    self.message = response.json()
                
                #additional parsing will happen in the child object that called in to here.
                
            else:   #soap call
                self.code = response[0] #suds puts the code in tuple position 0
                body = response[1]  #and the result in tuple position 1

                # Store the Last Request ID for use with continue
                if body and 'RequestID' in body:
                    self.request_id = body['RequestID']

                if self.code == 200:
                    self.status = True

                    if 'OverallStatus' in body:
                        self.message = body['OverallStatus']
                        if body['OverallStatus'] == "MoreDataAvailable":
                            self.more_results = True
                        elif body['OverallStatus'] != "OK":
                            self.status = False 
        
                    body_container_tag = None
                    if 'Results' in body:   #most SOAP responses are wrapped in 'Results'
                        body_container_tag = 'Results'
                    elif 'ObjectDefinition' in body:   #Describe SOAP response is in 'ObjectDefinition'
                        body_container_tag = 'ObjectDefinition'
                        
                    if body_container_tag is not None:
                        self.results = body[body_container_tag]                 

                else:
                    self.status = False
                    
    def parse_props_dict_into_ws_object(self, obj_type, ws_object, props_dict):
        for k, v in props_dict.items():
            if k in ws_object:
                ws_object[k] = v
            else:
                message = k + ' is not a property of ' + obj_type
                print(message)
                raise Exception(message)
        return ws_object

    def parse_props_into_ws_object(self, auth_stub, obj_type, props):
        empty_obj = auth_stub.soap_client.factory.create(obj_type)
        if props is not None and type(props) is dict:
            ws_create = copy.copy(empty_obj)
            ws_create = self.parse_props_dict_into_ws_object(obj_type, ws_create, props)
            return ws_create
        elif props is not None and type(props) is list:
            ws_create_list = []
            for prop_dict in props:
                #~ print str(datetime.now())+" - start"
                ws_create = copy.copy(empty_obj)
                ws_create = self.parse_props_dict_into_ws_object(obj_type, ws_create, prop_dict)
                #~ print str(datetime.now())+" - start"
                ws_create_list.append(ws_create)
            return ws_create_list
        else:
            message = 'Can not post properties to ' + obj_type + ' without a dict or list of properties'
            print(message)
            raise Exception(message)

########
##
##  Used to Describe Objects via web service call
##
########
class ET_Describe(ET_Constructor):
    def __init__(self, auth_stub, obj_type):        
        auth_stub.refresh_token()

        ws_describeRequest = auth_stub.soap_client.factory.create('ArrayOfObjectDefinitionRequest')

        ObjectDefinitionRequest = { 'ObjectType' : obj_type}
        ws_describeRequest.ObjectDefinitionRequest = [ObjectDefinitionRequest]

        response = auth_stub.soap_client.service.Describe(ws_describeRequest)       

        if response is not None:
            self.message = 'Describe: ' + obj_type
            super(ET_Describe, self).__init__(response)

########
##
##    Used to Configure Objects via web service call
##
########
class ET_Configure(ET_Constructor):
    def __init__(self, auth_stub, obj_type, props = None, update = False, delete = False):
        auth_stub.refresh_token()

        ws_configureRequest = auth_stub.soap_client.factory.create('ConfigureRequestMsg')
        action = 'create'
        if delete:
            action = 'delete'
        elif update:
            action = 'update'
        ws_configureRequest.Action = action
        ws_configureRequest.Configurations = {'Configuration': self.parse_props_into_ws_object(auth_stub, obj_type, props)}

        response = auth_stub.soap_client.service.Configure(None, ws_configureRequest)        

        if response is not None:
            #self.message = 'Describe: ' + obj_type
            super(ET_Configure, self).__init__(response)

########
##
##  Get call to a web service
##
########
class ET_Get(ET_Constructor):
    def __init__(self, auth_stub, obj_type, props = None, search_filter = None, options = None):        
        auth_stub.refresh_token()
        
        if props is None:   #if there are no properties to retrieve for the obj_type then return a Description of obj_type
            describe = ET_Describe(auth_stub, obj_type)
            props = []
            for prop in describe.results[0].Properties:
                if prop.IsRetrievable:
                    props.append(prop.Name) 

        ws_retrieveRequest = auth_stub.soap_client.factory.create('RetrieveRequest')
                
        if props is not None:
            if type(props) is dict: # If the properties is a hash, then we just want to use the keys
                ws_retrieveRequest.Properties = list(props.keys())
            else:
                ws_retrieveRequest.Properties = props

        if search_filter is not None:
            if 'LogicalOperator' in search_filter:
                ws_simpleFilterPartLeft = auth_stub.soap_client.factory.create('SimpleFilterPart')
                for prop in ws_simpleFilterPartLeft:
                    #print prop[0]
                    if prop[0] in search_filter['LeftOperand']:         
                        ws_simpleFilterPartLeft[prop[0]] = search_filter['LeftOperand'][prop[0]]    
                        
                ws_simpleFilterPartRight = auth_stub.soap_client.factory.create('SimpleFilterPart')
                for prop in ws_simpleFilterPartRight:
                    if prop[0] in search_filter['RightOperand']:
                        ws_simpleFilterPartRight[prop[0]] = search_filter['RightOperand'][prop[0]]
                        
                ws_complexFilterPart = auth_stub.soap_client.factory.create('ComplexFilterPart')
                ws_complexFilterPart.LeftOperand = ws_simpleFilterPartLeft
                ws_complexFilterPart.RightOperand = ws_simpleFilterPartRight
                ws_complexFilterPart.LogicalOperator = search_filter['LogicalOperator']
                for additional_operand in search_filter.get('AdditionalOperands', []):
                    ws_simpleFilterPart = auth_stub.soap_client.factory.create('SimpleFilterPart')
                    for k, v in list(additional_operand.items()):
                        ws_simpleFilterPart[k] = v
                    ws_complexFilterPart.AdditionalOperands.Operand.append(ws_simpleFilterPart)

                ws_retrieveRequest.Filter = ws_complexFilterPart
            else:
                ws_simpleFilterPart = auth_stub.soap_client.factory.create('SimpleFilterPart')
                for prop in ws_simpleFilterPart:
                    if prop[0] in search_filter:
                        ws_simpleFilterPart[prop[0]] = search_filter[prop[0]]
                ws_retrieveRequest.Filter = ws_simpleFilterPart

        if options is not None:
            for key, value in options.items():
                if isinstance(value, dict):
                    for k, v in value.items():
                        ws_retrieveRequest.Options[key][k] = v
                else:
                    ws_retrieveRequest.Options[key] = value

        ws_retrieveRequest.ObjectType = obj_type
        
        response = auth_stub.soap_client.service.Retrieve(ws_retrieveRequest)       

        if response is not None:
            super(ET_Get, self).__init__(response)

########
##
##  Call the Exact Target web service Create method
##
########
class ET_Post(ET_Constructor):
    def __init__(self, auth_stub, obj_type, props = None):
        auth_stub.refresh_token()

        response = auth_stub.soap_client.service.Create(None, self.parse_props_into_ws_object(auth_stub, obj_type, props))
        if(response is not None):
            super(ET_Post, self).__init__(response)

########
##
##  Call the Exact Target web service Update method
##
########
class ET_Patch(ET_Constructor):
    def __init__(self, auth_stub, obj_type, props = None):
        auth_stub.refresh_token()
              
        response = auth_stub.soap_client.service.Update(None, self.parse_props_into_ws_object(auth_stub, obj_type, props))

        if(response is not None):
            super(ET_Patch, self).__init__(response)

########
##
##  Call the Exact Target web service Delete method
##
########
class ET_Delete(ET_Constructor):
    def __init__(self, auth_stub, obj_type, props = None):
        auth_stub.refresh_token()
              
        response = auth_stub.soap_client.service.Delete(None, self.parse_props_into_ws_object(auth_stub, obj_type, props))

        if(response is not None):
            super(ET_Delete, self).__init__(response)

########
##
##  Call the Exact Target web service RetrieveRequest passing in ContinueRequest param
##
########
class ET_Continue(ET_Constructor):
    def __init__(self, auth_stub, request_id):
        auth_stub.refresh_token()

        ws_continueRequest = auth_stub.soap_client.factory.create('RetrieveRequest')
        ws_continueRequest.ContinueRequest = request_id
        response = auth_stub.soap_client.service.Retrieve(ws_continueRequest)       

        if response is not None:
            super(ET_Continue, self).__init__(response)

########
##
##  set up variables for children objects to share
##
########
class ET_BaseObject(object):
    auth_stub = None
    obj = None
    last_request_id = None
    endpoint = None
    props = None
    extProps = None
    search_filter = None
    options = None

########
##
##  make sure needed information is available and then make the call to ET_Get to call the webservice
##
########
class ET_GetSupport(ET_BaseObject):
    obj_type = 'ET_GetSupport'   #should be overwritten by inherited class
    
    def get(self, m_props = None, m_filter = None, m_options = None):
        props = self.props
        search_filter = self.search_filter
        options = self.options
        
        if m_props is not None and type(m_props) is list:
            props = m_props     
        elif self.props is not None and type(self.props) is dict:
            props = list(self.props.keys())

        if m_filter is not None and type(m_filter) is dict:
            search_filter = m_filter

        if m_options is not None and type(m_filter) is dict:
            options = m_options

        obj = ET_Get(self.auth_stub, self.obj_type, props, search_filter, options)
        if obj is not None:
            self.last_request_id = obj.request_id
        return obj
    
    def info(self):
        obj = ET_Describe(self.auth_stub, self.obj_type)
        if obj is not None:
            self.last_request_id = obj.request_id
        return obj
    
    def getMoreResults(self):
        obj = ET_Continue(self.auth_stub, self.last_request_id)
        if obj is not None:
            self.last_request_id = obj.request_id
        return obj

########
##
##  Restful webservice to Get data
##
########
class ET_GetRest(ET_Constructor):
    def __init__(self, auth_stub, endpoint, qs = None):
        auth_stub.refresh_token()   
        fullendpoint = endpoint + '?access_token=' + auth_stub.authToken
        for qStringValue in qs:
            fullendpoint += '&'+    qStringValue + '=' + str(qs[qStringValue])

        r = requests.get(fullendpoint)
    
        
        self.more_results = False
                    
        obj = super(ET_GetRest, self).__init__(r, True)
        return obj

########
##
##  Restful webservice to Get data
##
########
class ET_PostRest(ET_Constructor):  
    def __init__(self, auth_stub, endpoint, payload):
        auth_stub.refresh_token()
        
        headers = {'content-type' : 'application/json', 'user-agent' : 'FuelSDK-Python'}
        r = requests.post(endpoint + '?access_token=' + auth_stub.authToken , data=json.dumps(payload), headers=headers)
        
        obj = super(ET_PostRest, self).__init__(r, True)
        return obj
    
########
##
##  Restful webservice to Get data
##
########
class ET_PatchRest(ET_Constructor):
    def __init__(self, auth_stub, endpoint, payload):
        auth_stub.refresh_token()
        
        headers = {'content-type' : 'application/json', 'user-agent' : 'FuelSDK-Python'}
        r = requests.patch(endpoint + '?access_token=' + auth_stub.authToken , data=json.dumps(payload), headers=headers)
        
        obj = super(ET_PatchRest, self).__init__(r, True)
        return obj

########
##
##  Restful webservice to Get data
##
########
class ET_DeleteRest(ET_Constructor):
    def __init__(self, auth_stub, endpoint):
        auth_stub.refresh_token()
        
        r = requests.delete(endpoint + '?access_token=' + auth_stub.authToken)
        
        obj = super(ET_DeleteRest, self).__init__(r, True)
        return obj

########
##
##  Get data
##
########
class ET_CUDSupport(ET_GetSupport):
    
    def __init__(self):
        super(ET_CUDSupport, self).__init__()
        
    def post(self):
        if self.extProps is not None:
            for k, v in self.extProps.items():
                self.props[k.capitalize] = v
        
        obj = ET_Post(self.auth_stub, self.obj_type, self.props)
        if obj is not None:
            self.last_request_id = obj.request_id
        return obj
    
    def patch(self):
        obj = ET_Patch(self.auth_stub, self.obj_type, self.props)
        if obj is not None:
            self.last_request_id = obj.request_id
        return obj

    def delete(self):
        obj = ET_Delete(self.auth_stub, self.obj_type, self.props)
        if obj is not None:
            self.last_request_id = obj.request_id
        return obj

########
##
##  Get data using a REST call
##
########
class ET_GetSupportRest(ET_BaseObject):
    urlProps = None
    urlPropsRequired = None
    lastPageNumber = None
    
    def __init__(self):
        super
    
    def get(self, props = None):
        if props is not None and type(props) is dict:
            self.props = props
            
        completeURL = self.endpoint     
        additionalQS = {}
        
        if self.props is not None and type(self.props) is dict:
            for k, v in self.props.items():
                if k in self.urlProps:
                    completeURL = completeURL.replace('{{{0}}}'.format(k), v)
                else:
                    additionalQS[k] = v
        
        for value in self.urlPropsRequired: 
            if self.props is None or value not in self.props:
                raise "Unable to process request due to missing required prop: #{value}"
        
        for value in self.urlProps:          
            completeURL = completeURL.replace('/{{{0}}}'.format(value), '')

        obj = ET_GetRest(self.auth_stub, completeURL, additionalQS) 
        
        results = obj.results
        if 'page' in obj.results: 
            self.lastPageNumber = obj.results['page']
            pageSize = obj.results['pageSize']
            if 'count' in obj.results: 
                count = obj.results['count']
            elif 'totalCount' in obj.results: 
                count = obj.results['totalCount']
                    
            if count is not None and count > (self.lastPageNumber * pageSize):
                obj.more_results = True
            else:
                obj.more_results = False            
        return obj
    
    def getMoreResults(self):
        props = None    #where should it come from?
        if props is not None and type(props) is dict:
            self.props = props
        
        originalPageValue = "1"
        removePageFromProps = False         
        
        if self.props is not None and '$page' in self.props: 
            originalPageValue = self.props['page']
        else:
            removePageFromProps = True 
        
        if self.props is None:
            self.props = {}
        
        self.props['$page'] = self.lastPageNumber + 1
        
        obj = self.get()
        
        if removePageFromProps:
            del self.props['$page']
        else:
            self.props['$page'] = originalPageValue
        
        return obj

########
##
##  Create, Update and Delete using a REST call
##
########            
class ET_CUDSupportRest(ET_GetSupportRest):
    endpoint = None
    urlProps = None
    urlPropsRequired = None
    
    def __init__(self):
        super
    
    def post(self):
        completeURL = self.endpoint 
        
        if self.props is not None and type(self.props) is dict:
            for k, v in self.props.items():
                if k in self.urlProps:
                    completeURL = completeURL.replace('{{{0}}}'.format(k), v)
        
        for value in self.urlPropsRequired: 
            if self.props is None or value not in self.props:
                raise "Unable to process request due to missing required prop: #{value}"

        # Clean Optional Parameters from Endpoint URL first 
        for value in self.urlProps:          
            completeURL = completeURL.replace('/{{{0}}}'.format(value), '')      

        obj = ET_PostRest(self.auth_stub, completeURL, self.props)
        return obj      
    
    def patch(self):
        completeURL = self.endpoint
        # All URL Props are required when doing Patch   
        for value in self.urlProps: 
            if self.props is None or value not in self.props:
                raise "Unable to process request due to missing required prop: #{value}"
        
        if self.props is not None and type(self.props) is dict:
            for k, v in self.props.items():
                if k in self.urlProps:
                    completeURL = completeURL.replace('{{{0}}}'.format(k), v)
        
        obj = ET_PatchRest(self.auth_stub, completeURL, self.props)         
        return obj
    
    def delete(self):
        completeURL = self.endpoint     
        # All URL Props are required when doing Patch   
        for value in self.urlProps: 
            if self.props is None or value not in self.props:
                raise "Unable to process request due to missing required prop: #{value}"
        
        if self.props is not None and type(self.props) is dict:     
            for k, v in self.props.items():
                if k in self.urlProps:
                    completeURL = completeURL.replace('{{{0}}}'.format(k), v)

        obj = ET_DeleteRest(self.auth_stub, completeURL)
        return obj



__version__ = '0.9.4'

# Runtime patch the suds library
from .suds_patch import _PropertyAppender
from suds.mx import appender as _appender
_appender.PropertyAppender = _PropertyAppender

from .suds_patch import _bodycontent
from suds.bindings import document as _document
_document.Document.bodycontent = _bodycontent
# end runtime patching of suds

from .client import ET_Client

# Import all the wrapper objects
from .objects import *



import datetime
import suds


def partition_all(collection, chunk_size):
    return (collection[index:(index + chunk_size)]
            for index
            in range(0, len(collection), chunk_size))


def sudsobj_to_dict(obj):
    if isinstance(obj, list):
        return [sudsobj_to_dict(item) for item in obj]

    if not isinstance(obj, suds.sudsobject.Object):
        if isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%dT%H:%M:%SZ')

        return obj

    to_return = {}

    for key in obj.__keylist__:
        to_return[key] = sudsobj_to_dict(getattr(obj, key))

    return to_return



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.schemas import ID_FIELD, CUSTOM_PROPERTY_LIST, \
    CREATED_DATE_FIELD, CUSTOMER_KEY_FIELD, OBJECT_ID_FIELD, \
    MODIFIED_DATE_FIELD, with_properties
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table

LOGGER = singer.get_logger()


class ListSendDataAccessObject(DataAccessObject):
    SCHEMA = with_properties({
        'CreatedDate': CREATED_DATE_FIELD,
        'CustomerKey': CUSTOMER_KEY_FIELD,
        'ExistingUndeliverables': {
            'type': ['null', 'integer'],
            'description': ('Indicates whether bounces occurred on previous '
                            'send.'),
        },
        'ExistingUnsubscribes': {
            'type': ['null', 'integer'],
            'description': ('Indicates whether unsubscriptions occurred on '
                            'previous send.'),
        },
        'ForwardedEmails': {
            'type': ['null', 'integer'],
            'description': ('Number of emails forwarded for a send.'),
        },
        'HardBounces': {
            'type': ['null', 'integer'],
            'description': ('Indicates number of hard bounces associated '
                            'with a send.'),
        },
        'InvalidAddresses': {
            'type': ['null', 'integer'],
            'description': ('Specifies the number of invalid addresses '
                            'associated with a send.'),
        },
        'ListID': {
            'type': ['null', 'integer'],
            'description': 'List associated with the send.',
        },
        'ID': ID_FIELD,
        'MissingAddresses': {
            'type': ['null', 'integer'],
            'description': ('Specifies number of missing addresses '
                            'encountered within a send.'),
        },
        'ModifiedDate': MODIFIED_DATE_FIELD,
        'NumberDelivered': {
            'type': ['null', 'integer'],
            'description': ('Number of sent emails that did not bounce.'),
        },
        'NumberSent': {
            'type': ['null', 'integer'],
            'description': ('Number of emails actually sent as part of an '
                            'email send. This number reflects all of the sent '
                            'messages and may include bounced messages.'),
        },
        'ObjectID': OBJECT_ID_FIELD,
        'OtherBounces': {
            'type': ['null', 'integer'],
            'description': 'Specifies number of Other-type bounces in a send.',
        },
        'PartnerProperties': CUSTOM_PROPERTY_LIST,
        'SendID': {
            'type': ['null', 'integer'],
            'description': 'Contains identifier for a specific send.',
        },
        'SoftBounces': {
            'type': ['null', 'integer'],
            'description': ('Indicates number of soft bounces associated with '
                            'a specific send.'),
        },
        'UniqueClicks': {
            'type': ['null', 'integer'],
            'description': 'Indicates number of unique clicks on message.',
        },
        'UniqueOpens': {
            'type': ['null', 'integer'],
            'description': ('Indicates number of unique opens resulting from '
                            'a triggered send.'),
        },
        'Unsubscribes': {
            'type': ['null', 'integer'],
            'description': ('Indicates the number of unsubscribe events '
                            'associated with a send.'),
        },
    })

    TABLE = 'list_send'
    KEY_PROPERTIES = ['ListID', 'SendID']

    def parse_object(self, obj):
        to_return = obj.copy()

        to_return['ListID'] = to_return.get('List', {}).get('ID')

        return super(ListSendDataAccessObject, self).parse_object(to_return)

    def sync_data(self):
        table = self.__class__.TABLE
        selector = FuelSDK.ET_ListSend

        search_filter = None
        retrieve_all_since = get_last_record_value_for_table(self.state, table)

        if retrieve_all_since is not None:
            search_filter = {
                'Property': 'ModifiedDate',
                'SimpleOperator': 'greaterThan',
                'Value': retrieve_all_since
            }

        stream = request('ListSend',
                         selector,
                         self.auth_stub,
                         search_filter)

        for list_send in stream:
            list_send = self.filter_keys_and_parse(list_send)

            self.state = incorporate(self.state,
                                     table,
                                     'ModifiedDate',
                                     list_send.get('ModifiedDate'))

            singer.write_records(table, [list_send])

        save_state(self.state)



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.pagination import get_date_page, before_today, \
    increment_date
from tap_exacttarget.schemas import SUBSCRIBER_KEY_FIELD, with_properties
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table


LOGGER = singer.get_logger()


class EventDataAccessObject(DataAccessObject):
    SCHEMA = with_properties({
        'SendID': {
            'type': 'integer',
            'description': 'Contains identifier for a specific send.',
        },
        'EventDate': {
            'type': 'string',
            'format': 'datetime',
            'description': 'Date when a tracking event occurred.',
        },
        'EventType': {
            'type': 'string',
            'description': 'The type of tracking event',
        },
        'SubscriberKey': SUBSCRIBER_KEY_FIELD,
    })

    TABLE = 'event'
    KEY_PROPERTIES = ['SendID', 'EventType', 'SubscriberKey', 'EventDate']

    def sync_data(self):
        table = self.__class__.TABLE
        endpoints = {
            'sent': FuelSDK.ET_SentEvent,
            'click': FuelSDK.ET_ClickEvent,
            'open': FuelSDK.ET_OpenEvent,
            'bounce': FuelSDK.ET_BounceEvent,
            'unsub': FuelSDK.ET_UnsubEvent
        }

        for event_name, selector in endpoints.items():
            search_filter = None

            start = get_last_record_value_for_table(self.state, event_name)

            if start is None:
                start = self.config.get('start_date')

            if start is None:
                raise RuntimeError('start_date not defined!')

            pagination_unit = self.config.get(
                'pagination__{}_interval_unit'.format(event_name), 'minutes')
            pagination_quantity = self.config.get(
                'pagination__{}_interval_quantity'.format(event_name), 10)

            unit = {pagination_unit: int(pagination_quantity)}

            end = increment_date(start, unit)

            while before_today(start):
                LOGGER.info("Fetching {} from {} to {}"
                            .format(event_name, start, end))

                search_filter = get_date_page('EventDate', start, unit)

                stream = request(event_name,
                                 selector,
                                 self.auth_stub,
                                 search_filter)

                for event in stream:
                    event = self.filter_keys_and_parse(event)

                    self.state = incorporate(self.state,
                                             event_name,
                                             'EventDate',
                                             event.get('EventDate'))

                    singer.write_records(table, [event])

                self.state = incorporate(self.state,
                                         event_name,
                                         'EventDate',
                                         start)

                save_state(self.state)

                start = end
                end = increment_date(start, unit)



import singer

from funcy import project

from tap_exacttarget.util import sudsobj_to_dict

LOGGER = singer.get_logger()


def _get_catalog_schema(catalog):
    return catalog.get('schema', {}).get('properties')


class DataAccessObject(object):

    def __init__(self, config, state, auth_stub, catalog):
        self.config = config.copy()
        self.state = state.copy()
        self.catalog = catalog
        self.auth_stub = auth_stub

    @classmethod
    def matches_catalog(cls, catalog):
        return catalog.get('stream') == cls.TABLE

    def generate_catalog(self):
        cls = self.__class__

        return [{
            'tap_stream_id': cls.TABLE,
            'stream': cls.TABLE,
            'key_properties': cls.KEY_PROPERTIES,
            'schema': cls.SCHEMA,
        }]

    def filter_keys_and_parse(self, obj):
        to_return = sudsobj_to_dict(obj)

        return self.parse_object(to_return)

    def get_catalog_keys(self):
        return list(
            self.catalog.get('schema', {}).get('properties', {}).keys())

    def parse_object(self, obj):
        return project(obj, self.get_catalog_keys())

    def write_schema(self):
        singer.write_schema(
            self.catalog.get('stream'),
            self.catalog.get('schema'),
            key_properties=self.catalog.get('key_properties'))

    def sync(self):
        default = self.catalog.get('schema', {}) \
                              .get('selected-by-default', False)

        if not self.catalog.get('schema', {}).get('selected', default):
            LOGGER.info('{} is not marked as selected, skipping.'
                        .format(self.catalog.get('stream')))
            return

        LOGGER.info('Syncing stream {} with accessor {}'
                    .format(self.catalog.get('tap_stream_id'),
                            self.__class__.__name__))

        self.write_schema()

        return self.sync_data()

    # OVERRIDE THESE TO IMPLEMENT A NEW DAO:

    SCHEMA = None
    TABLE = None
    KEY_PROPERTIES = None

    def sync_data(self):  # pylint: disable=no-self-use
        raise RuntimeError('sync_data is not implemented!')



import unittest

from tap_exacttarget.pagination import increment_date


class TestPagination(unittest.TestCase):

    def test_increment_date(self):
        self.assertEqual(
            increment_date("2015-09-28T10:05:53Z"),
            "2015-09-29T10:05:53Z")
        self.assertEqual(
            increment_date("2015-09-28T10:05:53Z", {'hours': 1}),
            "2015-09-28T11:05:53Z")



import json
from dateutil.parser import parse

import singer

from voluptuous import Schema, Required

LOGGER = singer.get_logger()

STATE_SCHEMA = Schema({
    Required('bookmarks'): {
        str: {
            Required('last_record'): str,
            Required('field'): str,
        }
    }
})


def get_last_record_value_for_table(state, table):
    return state.get('bookmarks', {}) \
                .get(table, {}) \
                .get('value')


def incorporate(state, table, field, value):
    if value is None:
        return state

    new_state = state.copy()

    parsed = parse(value).strftime("%Y-%m-%dT%H:%M:%SZ")

    if 'bookmarks' not in new_state:
        new_state['bookmarks'] = {}

    if(new_state['bookmarks'].get(table, {}).get('last_record') is None or
       new_state['bookmarks'].get(table, {}).get('last_record') < value):
        new_state['bookmarks'][table] = {
            'field': field,
            'last_record': parsed,
        }

    return new_state


def save_state(state):
    if not state:
        return

    STATE_SCHEMA(state)

    LOGGER.info('Updating state.')

    singer.write_state(state)


def load_state(filename):
    if filename is None:
        return {}

    try:
        with open(filename) as handle:
            return json.load(handle)
    except:
        LOGGER.fatal("Failed to decode state file. Is it valid json?")
        raise RuntimeError



import os
import logging
import configparser
import time
import json

import jwt
import requests
import suds.client
import suds.wsse
from suds.sax.element import Element

from .objects import ET_DataExtension,ET_Subscriber


class ET_Client(object):
    """
    Setup web service connectivity by getting need config data, security tokens etc.
    """

    debug = False
    client_id = None
    client_secret = None
    appsignature = None
    wsdl_file_url = None
    authToken = None
    internalAuthToken = None
    authTokenExpiration = None  #seconds since epoch that the current jwt token will expire
    refreshKey = None
    endpoint = None
    authObj = None
    soap_client = None
    auth_url = None
        
    ## get_server_wsdl - if True and a newer WSDL is on the server than the local filesystem retrieve it
    def __init__(self, get_server_wsdl = False, debug = False, params = None):
        self.debug = debug
        if debug:
            logging.basicConfig(level=logging.INFO)
            logging.getLogger('suds.client').setLevel(logging.DEBUG)
            logging.getLogger('suds.transport').setLevel(logging.DEBUG)
            logging.getLogger('suds.xsd.schema').setLevel(logging.DEBUG)
            logging.getLogger('suds.wsdl').setLevel(logging.DEBUG)
        else:
            logging.getLogger('suds').setLevel(logging.INFO)

        ## Read the config information out of config.python
        config = configparser.RawConfigParser()
        if os.path.exists(os.path.expanduser('~/.fuelsdk/config.python')):
            config.read(os.path.expanduser('~/.fuelsdk/config.python'))
        else:
            config.read('config.python')

        if params is not None and 'clientid' in params:
            self.client_id = params['clientid']
        elif config.has_option('Web Services', 'clientid'):
            self.client_id = config.get('Web Services', 'clientid')
        elif 'FUELSDK_CLIENT_ID' in os.environ:
            self.client_id = os.environ['FUELSDK_CLIENT_ID']

        if params is not None and 'clientsecret' in params:
            self.client_secret = params['clientsecret']
        elif config.has_option('Web Services', 'clientsecret'):
            self.client_secret = config.get('Web Services', 'clientsecret')
        elif 'FUELSDK_CLIENT_SECRET' in os.environ:
            self.client_secret = os.environ['FUELSDK_CLIENT_SECRET']

        if params is not None and 'appsignature' in params:
            self.appsignature = params['appsignature']
        elif config.has_option('Web Services', 'appsignature'):
            self.appsignature = config.get('Web Services', 'appsignature')
        elif 'FUELSDK_APP_SIGNATURE' in os.environ:
            self.appsignature = os.environ['FUELSDK_APP_SIGNATURE']

        if params is not None and 'defaultwsdl' in params:
            wsdl_server_url = params['defaultwsdl']
        elif config.has_option('Web Services', 'defaultwsdl'):
            wsdl_server_url = config.get('Web Services', 'defaultwsdl')
        elif 'FUELSDK_DEFAULT_WSDL' in os.environ:
            wsdl_server_url = os.environ['FUELSDK_DEFAULT_WSDL']
        else:
            wsdl_server_url = 'https://webservice.exacttarget.com/etframework.wsdl'

        if params is not None and 'authenticationurl' in params:
            self.auth_url = params['authenticationurl']
        elif config.has_option('Web Services', 'authenticationurl'):
            self.auth_url = config.get('Web Services', 'authenticationurl')
        elif 'FUELSDK_AUTH_URL' in os.environ:
            self.auth_url = os.environ['FUELSDK_AUTH_URL']
        else:
            self.auth_url = 'https://auth.exacttargetapis.com/v1/requestToken?legacy=1'

        if params is not None and "wsdl_file_local_loc" in params:
            wsdl_file_local_location = params["wsdl_file_local_loc"]
        elif config.has_option("Web Services", "wsdl_file_local_loc"):
            wsdl_file_local_location = config.get("Web Services", "wsdl_file_local_loc")
        elif "FUELSDK_WSDL_FILE_LOCAL_LOC" in os.environ:
            wsdl_file_local_location = os.environ["FUELSDK_WSDL_FILE_LOCAL_LOC"]
        else:
            wsdl_file_local_location = None

        self.wsdl_file_url = self.load_wsdl(wsdl_server_url, wsdl_file_local_location, get_server_wsdl)

        ## get the JWT from the params if passed in...or go to the server to get it             
        if(params is not None and 'jwt' in params):
            decodedJWT = jwt.decode(params['jwt'], self.appsignature)
            self.authToken = decodedJWT['request']['user']['oauthToken']
            self.authTokenExpiration = time.time() + decodedJWT['request']['user']['expiresIn']
            self.internalAuthToken = decodedJWT['request']['user']['internalOauthToken']
            if 'refreshToken' in decodedJWT:
                self.refreshKey = tokenResponse['request']['user']['refreshToken']
            self.build_soap_client()
            pass
        else:
            self.refresh_token()


    def load_wsdl(self, wsdl_url, wsdl_file_local_location, get_server_wsdl = False):
        """
        retrieve the url of the ExactTarget wsdl...either file: or http:
        depending on if it already exists locally or server flag is set and
        server has a newer copy
        """
        if wsdl_file_local_location is not None:
           file_location = wsdl_file_local_location
        else:
           path = os.path.dirname(os.path.abspath(__file__))
           file_location = os.path.join(path, 'ExactTargetWSDL.xml')
        file_url = 'file:///' + file_location 
        
        if not os.path.exists(file_location) or os.path.getsize(file_location) == 0:   #if there is no local copy or local copy is empty then go get it...
            self.retrieve_server_wsdl(wsdl_url, file_location)
        elif get_server_wsdl:
            r = requests.head(wsdl_url)
            if r is not None and 'last-modified' in r.headers:
                server_wsdl_updated = time.strptime(r.headers['last-modified'], '%a, %d %b %Y %H:%M:%S %Z')
                file_wsdl_updated = time.gmtime(os.path.getmtime(file_location))
                if server_wsdl_updated > file_wsdl_updated:
                    self.retrieve_server_wsdl(wsdl_url, file_location)
            
        return file_url
            

    def retrieve_server_wsdl(self, wsdl_url, file_location):
        """
        get the WSDL from the server and save it locally
        """
        r = requests.get(wsdl_url)
        f = open(file_location, 'w')
        f.write(r.text)
        

    def build_soap_client(self):
        if self.endpoint is None: 
            self.endpoint = self.determineStack()
        
        self.authObj = {'oAuth' : {'oAuthToken' : self.internalAuthToken}}          
        self.authObj['attributes'] = { 'oAuth' : { 'xmlns' : 'http://exacttarget.com' }}                        

        self.soap_client = suds.client.Client(self.wsdl_file_url, faults=False, cachingpolicy=1)
        self.soap_client.set_options(location=self.endpoint)

        element_oAuth = Element('oAuth', ns=('etns', 'http://exacttarget.com'))
        element_oAuthToken = Element('oAuthToken').setText(self.internalAuthToken)
        element_oAuth.append(element_oAuthToken)
        self.soap_client.set_options(soapheaders=(element_oAuth))               
        
        security = suds.wsse.Security()
        token = suds.wsse.UsernameToken('*', '*')
        security.tokens.append(token)
        self.soap_client.set_options(wsse=security)             
        

    def refresh_token(self, force_refresh = False):
        """
        Called from many different places right before executing a SOAP call
        """
        #If we don't already have a token or the token expires within 5 min(300 seconds), get one
        if (force_refresh or self.authToken is None or (self.authTokenExpiration is not None and time.time() + 300 > self.authTokenExpiration)):
            headers = {'content-type' : 'application/json', 'user-agent' : 'FuelSDK-Python'}
            if (self.authToken is None):
                payload = {'clientId' : self.client_id, 'clientSecret' : self.client_secret, 'accessType': 'offline'}
            else:
                payload = {'clientId' : self.client_id, 'clientSecret' : self.client_secret, 'refreshToken' : self.refreshKey, 'accessType': 'offline'}
            if self.refreshKey:
                payload['refreshToken'] = self.refreshKey

            r = requests.post(self.auth_url, data=json.dumps(payload), headers=headers)
            tokenResponse = r.json()
            
            if 'accessToken' not in tokenResponse:
                raise Exception('Unable to validate App Keys(ClientID/ClientSecret) provided: ' + repr(r.json()))
            
            self.authToken = tokenResponse['accessToken']
            self.authTokenExpiration = time.time() + tokenResponse['expiresIn']
            self.internalAuthToken = tokenResponse['legacyToken']
            if 'refreshToken' in tokenResponse:
                self.refreshKey = tokenResponse['refreshToken']
        
            self.build_soap_client()
            

    def determineStack(self):
        """
        find the correct url that data request web calls should go against for the token we have.
        """
        try:
            r = requests.get('https://www.exacttargetapis.com/platform/v1/endpoints/soap?access_token=' + self.authToken, {'user-agent' : 'FuelSDK-Python'})
            contextResponse = r.json()
            if('url' in contextResponse):
                return str(contextResponse['url'])

        except Exception as e:
            raise Exception('Unable to determine stack using /platform/v1/tokenContext: ' + e.message)  


    def AddSubscriberToList(self, emailAddress, listIDs, subscriberKey = None):
        """
        add or update a subscriber with a list
        """
        newSub = ET_Subscriber()
        newSub.auth_stub = self
        lists = []
        
        for p in listIDs:
            lists.append({"ID" : p})
        
        newSub.props = {"EmailAddress" : emailAddress, "Lists" : lists}
        if subscriberKey is not None:
            newSub.props['SubscriberKey']  = subscriberKey
        
        # Try to add the subscriber
        postResponse = newSub.post()
        
        if not postResponse.status:
            # If the subscriber already exists in the account then we need to do an update.
            # Update Subscriber On List 
            if postResponse.results[0]['ErrorCode'] == 12014:    
                patchResponse = newSub.patch()
                return patchResponse

        return postResponse
    

    def CreateDataExtensions(self, dataExtensionDefinitions):
        """
        write the data extension props to the web service
        """
        newDEs = ET_DataExtension()
        newDEs.auth_stub = self
                
        newDEs.props = dataExtensionDefinitions                     
        postResponse = newDEs.post()        
        
        return postResponse



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.schemas import ID_FIELD, CUSTOM_PROPERTY_LIST, \
    CREATED_DATE_FIELD, CUSTOMER_KEY_FIELD, OBJECT_ID_FIELD, \
    MODIFIED_DATE_FIELD, with_properties
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table

LOGGER = singer.get_logger()


class EmailDataAccessObject(DataAccessObject):
    SCHEMA = with_properties({
        'CategoryID': {
            'type': ['null', 'integer'],
            'description': ('Specifies the identifier of the folder '
                            'containing the email.'),
        },
        'CharacterSet': {
            'type': ['null', 'string'],
            'description': ('Indicates encoding used in an email '
                            'message.'),
        },
        'ClonedFromID': {
            'type': ['null', 'integer'],
            'description': ('ID of email message from which the specified '
                            'email message was created.'),
        },
        'ContentAreaIDs': {
            'type': 'array',
            'description': ('Contains information on content areas '
                            'included in an email message.'),
            'items': {
                'type': 'integer'
            }
        },
        'ContentCheckStatus': {
            'type': ['null', 'string'],
            'description': ('Indicates whether content validation has '
                            'completed for this email message.'),
        },
        'CreatedDate': CREATED_DATE_FIELD,
        'CustomerKey': CUSTOMER_KEY_FIELD,
        'EmailType': {
            'type': ['null', 'string'],
            'description': ('Defines preferred email type.'),
        },
        'HasDynamicSubjectLine': {
            'type': 'boolean',
            'description': ('Indicates whether email message contains '
                            'a dynamic subject line.'),
        },
        'HTMLBody': {
            'type': ['null', 'string'],
            'description': ('Contains HTML body of an email message.'),
        },
        'ID': ID_FIELD,
        'IsActive': {
            'type': 'boolean',
            'description': ('Specifies whether the object is active.')
        },
        'IsHTMLPaste': {
            'type': 'boolean',
            'description': ('Indicates whether email message was created '
                            'via pasted HTML.')
        },
        'ModifiedDate': MODIFIED_DATE_FIELD,
        'Name': {
            'type': ['null', 'string'],
            'description': 'Name of the object or property.',
        },
        'ObjectID': OBJECT_ID_FIELD,
        'PartnerProperties': CUSTOM_PROPERTY_LIST,
        'PreHeader': {
            'type': ['null', 'string'],
            'description': ('Contains text used in preheader of email '
                            'message on mobile devices.')
        },
        'Status': {
            'type': ['null', 'string'],
            'description': ('Defines status of object. Status of an '
                            'address.'),
        },
        'Subject': {
            'type': ['null', 'string'],
            'description': ('Contains subject area information for a '
                            'message.'),
        },
        'SyncTextWithHTML': {
            'type': 'boolean',
            'description': ('Makes the text version of an email contain '
                            'the same content as the HTML version.'),
        },
        'TextBody': {
            'type': ['null', 'string'],
            'description': ('Contains raw text body of a message.'),
        },
        '__AdditionalEmailAttribute1': {'type': ['null', 'string']},
        '__AdditionalEmailAttribute2': {'type': ['null', 'string']},
        '__AdditionalEmailAttribute3': {'type': ['null', 'string']},
        '__AdditionalEmailAttribute4': {'type': ['null', 'string']},
        '__AdditionalEmailAttribute5': {'type': ['null', 'string']},
    })

    TABLE = 'email'
    KEY_PROPERTIES = ['ID']

    def parse_object(self, obj):
        to_return = obj.copy()
        content_area_ids = []

        for content_area in to_return.get('ContentAreas', []):
            content_area_ids.append(content_area.get('ID'))

        to_return['EmailID'] = to_return.get('Email', {}).get('ID')
        to_return['ContentAreaIDs'] = content_area_ids

        return super(EmailDataAccessObject, self).parse_object(to_return)

    def sync_data(self):
        table = self.__class__.TABLE
        selector = FuelSDK.ET_Email

        search_filter = None
        retrieve_all_since = get_last_record_value_for_table(self.state, table)

        if retrieve_all_since is not None:
            search_filter = {
                'Property': 'ModifiedDate',
                'SimpleOperator': 'greaterThan',
                'Value': retrieve_all_since
            }

        stream = request('Email',
                         selector,
                         self.auth_stub,
                         search_filter)

        for email in stream:
            email = self.filter_keys_and_parse(email)

            self.state = incorporate(self.state,
                                     table,
                                     'ModifiedDate',
                                     email.get('ModifiedDate'))

            singer.write_records(table, [email])

        save_state(self.state)



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.schemas import ID_FIELD, CUSTOM_PROPERTY_LIST, \
    CREATED_DATE_FIELD, OBJECT_ID_FIELD, DESCRIPTION_FIELD, \
    MODIFIED_DATE_FIELD, with_properties
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table


LOGGER = singer.get_logger()


class ListDataAccessObject(DataAccessObject):

    SCHEMA = with_properties({
        'Category': {
            'type': 'integer',
            'description': 'ID of the folder that an item is located in.',
        },
        'CreatedDate': CREATED_DATE_FIELD,
        'ID': ID_FIELD,
        'ModifiedDate': MODIFIED_DATE_FIELD,
        'ObjectID': OBJECT_ID_FIELD,
        'PartnerProperties': CUSTOM_PROPERTY_LIST,
        'ListClassification': {
            'type': ['null', 'string'],
            'description': ('Specifies the classification for a list.'),
        },
        'ListName': {
            'type': ['null', 'string'],
            'description': 'Name of a specific list.',
        },
        'Description': DESCRIPTION_FIELD,
        'SendClassification': {
            'type': ['null', 'string'],
            'description': ('Indicates the send classification to use '
                            'as part of a send definition.'),
        },
        'Type': {
            'type': ['null', 'string'],
            'description': ('Indicates type of specific list. Valid '
                            'values include Public, Private, Salesforce, '
                            'GlobalUnsubscribe, and Master.')
        }
    })

    TABLE = 'list'
    KEY_PROPERTIES = ['ID']

    def sync_data(self):
        table = self.__class__.TABLE
        selector = FuelSDK.ET_List

        search_filter = None
        retrieve_all_since = get_last_record_value_for_table(self.state, table)

        if retrieve_all_since is not None:
            search_filter = {
                'Property': 'ModifiedDate',
                'SimpleOperator': 'greaterThan',
                'Value': retrieve_all_since
            }

        stream = request('List',
                         selector,
                         self.auth_stub,
                         search_filter)

        for _list in stream:
            _list = self.filter_keys_and_parse(_list)

            self.state = incorporate(self.state,
                                     table,
                                     'ModifiedDate',
                                     _list.get('ModifiedDate'))

            singer.write_records(table, [_list])

        save_state(self.state)



from .rest import ET_CUDSupport,ET_CUDSupportRest,ET_GetSupport,ET_Get,ET_Patch,ET_Post,ET_Delete,ET_Configure

########
##
##  wrap an Exact Target Content Area
##
########
class ET_ContentArea(ET_CUDSupport):    
    def __init__(self):
        super(ET_ContentArea, self).__init__()
        self.obj_type = 'ContentArea'

########
##
##  wrap an Exact Target DataFolder
##
########
class ET_Folder(ET_CUDSupport): 
    def __init__(self):
        super(ET_Folder, self).__init__()
        self.obj_type = 'DataFolder'

########
##
##    wrap an Exact Target PropertyDefinition
##
########
class ET_ProfileAttribute():    
    def __init__(self):
        self.obj_type = 'PropertyDefinition'
        self.update = False
        self.delete = False

    def post(self):       
        obj = ET_Configure(self.auth_stub, self.obj_type, self.props, self.update, self.delete)
        if obj is not None:
            self.last_request_id = obj.request_id
        return obj

########
##
##  wrap an Exact Target Bounce Event
##
########
class ET_BounceEvent(ET_GetSupport):
    def __init__(self):
        self.obj_type = 'BounceEvent'
     
########
##
##  wrap an Exact Target Campaign and associated Assets
##
########        
class ET_Campaign(ET_CUDSupportRest):
    def __init__(self):
        super(ET_Campaign, self).__init__()
        self.endpoint = 'https://www.exacttargetapis.com/hub/v1/campaigns/{id}'
        self.urlProps = ["id"]
        self.urlPropsRequired = []
        
    ##the patch rest service is not implemented for campaigns yet.  use post instead and remove this when patch is implemented on the back end
    def patch(self):
        self.endpoint = 'https://www.exacttargetapis.com/hub/v1/campaigns'  #don't put the id on the url when patching via post
        obj = super(ET_Campaign, self).post()
        self.endpoint = 'https://www.exacttargetapis.com/hub/v1/campaigns/{id}' #but set it back to the url with id for other operations to continue working
        return obj
    
class ET_Campaign_Asset(ET_CUDSupportRest):
    def __init__(self):
        super(ET_Campaign_Asset, self).__init__()
        self.endpoint = 'https://www.exacttargetapis.com/hub/v1/campaigns/{id}/assets/{assetId}'
        self.urlProps = ["id", "assetId"]
        self.urlPropsRequired = ["id"]
        
########
##
##  wrap an Exact Target Click Event
##
########
class ET_ClickEvent(ET_GetSupport):
    def __init__(self):
        super(ET_ClickEvent, self).__init__()
        self.obj_type = 'ClickEvent'
        
########
##
##  wrap an Exact Target List and List Subscriber
##
########
class ET_Group(ET_CUDSupport):
    def __init__(self):
        super(ET_Group, self).__init__()
        self.obj_type = 'Group'

class ET_Send(ET_CUDSupport):
    def __init__(self):
        super(ET_Send, self).__init__()
        self.obj_type = 'Send'

class ET_ListSend(ET_CUDSupport):
    def __init__(self):
        super(ET_ListSend, self).__init__()
        self.obj_type = 'ListSend'

class ET_List(ET_CUDSupport):
    def __init__(self):
        super(ET_List, self).__init__()
        self.obj_type = 'List'

class ET_List_Subscriber(ET_GetSupport):
    def __init__(self):
        super(ET_List_Subscriber, self).__init__()
        self.obj_type = 'ListSubscriber'

class ET_SubscriberList(ET_GetSupport):
    def __init__(self):
        super(ET_SubscriberList, self).__init__()
        self.obj_type = 'SubscriberList'

class ET_SentEvent(ET_GetSupport):
    def __init__(self):
        super(ET_SentEvent, self).__init__()
        self.obj_type = 'SentEvent'

class ET_OpenEvent(ET_GetSupport):
    def __init__(self):
        super(ET_OpenEvent, self).__init__()
        self.obj_type = 'OpenEvent'

class ET_UnsubEvent(ET_GetSupport):
    def __init__(self):
        super(ET_UnsubEvent, self).__init__()
        self.obj_type = 'UnsubEvent'

class ET_Email(ET_CUDSupport):
    def __init__(self):
        super(ET_Email, self).__init__()
        self.obj_type = 'Email'

class ET_TriggeredSend(ET_CUDSupport):
    subscribers = None
    attributes  = None
    def __init__(self):
        super(ET_TriggeredSend, self).__init__()
        self.obj_type = 'TriggeredSendDefinition'

    def send(self):
        tscall = {"TriggeredSendDefinition" : self.props, "Subscribers" : self.subscribers, "Attributes": self.attributes }
        self.obj = ET_Post(self.auth_stub, "TriggeredSend", tscall)
        return self.obj


class ET_Subscriber(ET_CUDSupport):
    def __init__(self):
        super(ET_Subscriber, self).__init__()
        self.obj_type = 'Subscriber'
        
class ET_DataExtension(ET_CUDSupport):
    columns = None
    
    def __init__(self):
        super(ET_DataExtension, self).__init__()
        self.obj_type = 'DataExtension' 
    
    def post(self):
        originalProps = self.props

        if type(self.props) is list:
            multiDE = []
            for currentDE in self.props:
                currentDE['Fields'] = {}
                currentDE['Fields']['Field'] = []               
                for key in currentDE['columns']:                    
                    currentDE['Fields']['Field'].append(key)
                del currentDE['columns']
                multiDE.append(currentDE.copy())
        
            self.props = multiDE
        else:
            self.props['Fields'] = {}
            self.props['Fields']['Field'] = []
            
            for key in self.columns:        
                self.props['Fields']['Field'].append(key)
        
        obj = super(ET_DataExtension, self).post()
        self.props = originalProps  
        return obj

    def patch(self):
        self.props['Fields'] = {}
        self.props['Fields']['Field'] = []
        for key in self.columns:
            self.props['Fields']['Field'].append(key)
        obj = super(ET_DataExtension, self).patch()
        del self.props["Fields"]         
        return obj
    
class ET_DataExtension_Column(ET_GetSupport):
    def __init__(self):
        super(ET_DataExtension_Column, self).__init__()
        self.obj = 'DataExtensionField'
        
    def get(self):
        '''
        if props and props.is_a? Array then
            @props = props
        end
        '''
        
        if self.props is not None and type(self.props) is dict:
            self.props = list(self.props.keys())

        '''
        if filter and filter.is_a? Hash then
            @filter = filter
        end
        '''
                
        '''             
        fixCustomerKey = False
        if filter and filter.is_a? Hash then
            @filter = filter
            if @filter.has_key?("Property") && @filter["Property"] == "CustomerKey" then
                @filter["Property"]  = "DataExtension.CustomerKey"
                fixCustomerKey = true 
            end 
        end
        '''
        
        obj = ET_Get(self.auth_stub, self.obj, self.props, self.search_filter)                      
        self.last_request_id = obj.request_id   
        
        ''' 
        if fixCustomerKey then
            @filter["Property"] = "CustomerKey"
        end 
        '''
            
        return obj

class ET_DataExtension_Row(ET_CUDSupport):
    Name = None
    CustomerKey = None      
                
    def __init__(self):                             
        super(ET_DataExtension_Row, self).__init__()
        self.obj_type = "DataExtensionObject"
        
    def get(self):
        self.getName()
        '''
        if props and props.is_a? Array then
            @props = props
        end
        '''
        
        if self.props is not None and type(self.props) is dict:
            self.props = list(self.props.keys())

        '''
        if filter and filter.is_a? Hash then
            @filter = filter
        end
        '''
            
        obj = ET_Get(self.auth_stub, "DataExtensionObject[{0}]".format(self.Name), self.props, self.search_filter)                      
        self.last_request_id = obj.request_id               
            
        return obj
        
    def post(self):
        self.getCustomerKey()
        originalProps = self.props
        
        if type(self.props) is list:
            currentPropList = []
            for rec in self.props:
                currentFields = []
                currentProp = {}
                
                for key, value in rec.items():
                    currentFields.append({"Name" : key, "Value" : value})
                
                currentProp['CustomerKey'] = self.CustomerKey
                currentProp['Properties'] = {}
                currentProp['Properties']['Property'] = currentFields
                
                currentPropList.append(currentProp)
            
            currentProp = currentPropList

        else:
            currentFields = []
            currentProp = {}
                
            for key, value in self.props.items():
                currentFields.append({"Name" : key, "Value" : value})

            currentProp['CustomerKey'] = self.CustomerKey
            currentProp['Properties'] = {}
            currentProp['Properties']['Property'] = currentFields   

        obj = ET_Post(self.auth_stub, self.obj_type, currentProp)   
        self.props = originalProps
        return obj
        
    def patch(self): 
        self.getCustomerKey()

        if type(self.props) is list:
            currentPropList = []
            for rec in self.props:
                currentFields = []
                currentProp = {}
                
                for key, value in rec.items():
                    currentFields.append({"Name" : key, "Value" : value})
                
                currentProp['CustomerKey'] = self.CustomerKey
                currentProp['Properties'] = {}
                currentProp['Properties']['Property'] = currentFields
                
                currentPropList.append(currentProp)
            
            currentProp = currentPropList
        else:
            currentFields = []
            currentProp = {}
            
            for key, value in self.props.items():
                currentFields.append({"Name" : key, "Value" : value})
            
            currentProp['CustomerKey'] = self.CustomerKey
            currentProp['Properties'] = {}
            currentProp['Properties']['Property'] = currentFields
            
        obj = ET_Patch(self.auth_stub, self.obj_type, currentProp)
        return obj
    
    def delete(self): 
        self.getCustomerKey()

        if type(self.props) is list:
            currentPropList = []
            for rec in self.props:
                currentFields = []
                currentProp = {}
                
                for key, value in rec.items():
                    currentFields.append({"Name" : key, "Value" : value})
                
                currentProp['CustomerKey'] = self.CustomerKey
                currentProp['Keys'] = {}
                currentProp['Keys']['Key'] = currentFields
                
                currentPropList.append(currentProp)
            
            currentProp = currentPropList
        else:
            currentFields = []
            currentProp = {}
                
            for key, value in self.props.items():
                currentFields.append({"Name" : key, "Value" : value})
    
            currentProp['CustomerKey'] = self.CustomerKey
            currentProp['Keys'] = {}
            currentProp['Keys']['Key'] = currentFields
            
        obj = ET_Delete(self.auth_stub, self.obj_type, currentProp)
        return obj
    
    def getCustomerKey(self):
        if self.CustomerKey is None:
            if self.Name is None:    
                raise Exception('Unable to process DataExtension::Row request due to CustomerKey and Name not being defined on ET_DatExtension::row')   
            else:
                de = ET_DataExtension()
                de.auth_stub = self.auth_stub
                de.props = ["Name","CustomerKey"]
                de.search_filter = {'Property' : 'CustomerKey','SimpleOperator' : 'equals','Value' : self.Name}
                getResponse = de.get()
                if getResponse.status and len(getResponse.results) == 1 and 'CustomerKey' in getResponse.results[0]: 
                    self.CustomerKey = getResponse.results[0]['CustomerKey']
                else:
                    raise Exception('Unable to process DataExtension::Row request due to unable to find DataExtension based on Name')
                    
                
    def getName(self):
        if self.Name is None:
            if self.CustomerKey is None:
                raise Exception('Unable to process DataExtension::Row request due to CustomerKey and Name not being defined on ET_DatExtension::row')   
            else:
                de = ET_DataExtension()
                de.auth_stub = self.auth_stub
                de.props = ["Name","CustomerKey"]
                de.search_filter = {'Property' : 'CustomerKey','SimpleOperator' : 'equals','Value' : self.CustomerKey}
                getResponse = de.get()
                if getResponse.status and len(getResponse.results) == 1 and 'Name' in getResponse.results[0]: 
                    self.Name = getResponse.results[0]['Name']
                else:
                    raise Exception('Unable to process DataExtension::Row request due to unable to find DataExtension based on CustomerKey')



def with_properties(properties):
    return {
        'type': 'object',
        'inclusion': 'available',
        'selected-by-default': False,
        'properties': properties
    }


CUSTOM_PROPERTY_LIST = {
    'type': 'array',
    'description': ('Specifies key-value pairs of properties associated with '
                    'an object.'),
    'items': {
        'type': 'object',
        'properties': {
            'Name': {'type': 'string'},
            'Value': {'type': ['null', 'string']},
        }
    }
}

ID_FIELD = {
    'type': 'integer',
    'description': ('Read-only legacy identifier for an object. Not '
                    'supported on all objects. Some objects use the '
                    'ObjectID property as the Marketing Cloud unique '
                    'ID.')
}

CREATED_DATE_FIELD = {
    'type': ['null', 'string'],
    'description': ('Read-only date and time of the object\'s'
                    'creation.'),
}

MODIFIED_DATE_FIELD = {
    'type': ['null', 'string'],
    'description': ('Indicates the last time object information '
                    'was modified.')
}

CUSTOMER_KEY_FIELD = {
    'type': ['null', 'string'],
    'description': ('User-supplied unique identifier for an '
                    'object within an object type (corresponds '
                    'to the external key assigned to an object '
                    'in the user interface).'),
}

OBJECT_ID_FIELD = {
    'type': ['null', 'string'],
    'description': ('System-controlled, read-only text string '
                    'identifier for object.'),
}

DESCRIPTION_FIELD = {
    'type': ['null', 'string'],
    'description': ('Describes and provides information regarding '
                    'the object.'),
}

SUBSCRIBER_KEY_FIELD = {
    'type': 'string',
    'description': 'Identification of a specific subscriber.',
}





def simple(field, operator, value):
    return {
        'Property': field,
        'SimpleOperator': operator,
        'Value': value
    }


def combine(left, right, operator):
    return {
        'LogicalOperator': operator,
        'LeftOperand': left,
        'RightOperand': right
    }


def between(field, start, end):
    return simple(
        field,
        'between',
        [start, end])



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.schemas import ID_FIELD, CUSTOM_PROPERTY_LIST, \
    CREATED_DATE_FIELD, MODIFIED_DATE_FIELD, with_properties
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table

LOGGER = singer.get_logger()


class SendDataAccessObject(DataAccessObject):
    SCHEMA = with_properties({
        'CreatedDate': CREATED_DATE_FIELD,
        'EmailID': {
            'type': 'integer',
            'description': ('Specifies the ID of an email message '
                            'associated with a send.'),
        },
        'EmailName': {
            'type': 'string',
            'description': ('Specifies the name of an email message '
                            'associated with a send.'),
        },
        'FromAddress': {
            'type': 'string',
            'description': ('Indicates From address associated with a '
                            'object. Deprecated for email send '
                            'definitions and triggered send '
                            'definitions.'),
        },
        'FromName': {
            'type': 'string',
            'description': ('Specifies the default email message From '
                            'Name. Deprecated for email send '
                            'definitions and triggered send '
                            'definitions.'),
        },
        'ID': ID_FIELD,
        'IsAlwaysOn': {
            'type': 'boolean',
            'description': ('Indicates whether the request can be '
                            'performed while the system is is '
                            'maintenance mode. A value of true '
                            'indicates the system will process the '
                            'request.'),
        },
        'IsMultipart': {
            'type': 'boolean',
            'description': ('Indicates whether the email is sent with '
                            'Multipart/MIME enabled.'),
        },
        'ModifiedDate': MODIFIED_DATE_FIELD,
        'PartnerProperties': CUSTOM_PROPERTY_LIST,
        'SendDate': {
            'type': 'string',
            'format': 'date-time',
            'description': ('Indicates the date on which a send '
                            'occurred. Set this value to have a CST '
                            '(Central Standard Time) value.'),
        },
        'SentDate': {
            'type': 'string',
            'format': 'date-time',
            'description': ('Indicates date on which a send took '
                            'place.'),
        },
        'Status': {
            'type': 'string',
            'description': ('Defines status of object. Status of an '
                            'address.'),
        },
        'Subject': {
            'type': 'string',
            'description': ('Contains subject area information for '
                            'a message.'),
        }
    })

    TABLE = 'send'
    KEY_PROPERTIES = ['ID']

    def parse_object(self, obj):
        to_return = obj.copy()

        to_return['EmailID'] = to_return.get('Email', {}).get('ID')

        return super(SendDataAccessObject, self).parse_object(to_return)

    def sync_data(self):
        table = self.__class__.TABLE
        selector = FuelSDK.ET_Send

        search_filter = None
        retrieve_all_since = get_last_record_value_for_table(self.state, table)

        if retrieve_all_since is not None:
            search_filter = {
                'Property': 'ModifiedDate',
                'SimpleOperator': 'greaterThan',
                'Value': retrieve_all_since
            }

        stream = request('Send',
                         selector,
                         self.auth_stub,
                         search_filter)

        for send in stream:
            send = self.filter_keys_and_parse(send)

            self.state = incorporate(self.state,
                                     table,
                                     'ModifiedDate',
                                     send.get('ModifiedDate'))

            singer.write_records(table, [send])

        save_state(self.state)



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.schemas import CUSTOM_PROPERTY_LIST, ID_FIELD, \
    CREATED_DATE_FIELD, CUSTOMER_KEY_FIELD, OBJECT_ID_FIELD, \
    SUBSCRIBER_KEY_FIELD, MODIFIED_DATE_FIELD, with_properties


LOGGER = singer.get_logger()

SCHEMA = with_properties({
    'Addresses': {
        'type': 'array',
        'description': ('Indicates addresses belonging to a subscriber, '
                        'used to create, retrieve, update or delete an '
                        'email or SMS Address for a given subscriber.'),
        'items': {
            'type': 'object',
            'properties': {
                'Address': {'type': 'string'},
                'AddressType': {'type': 'string'},
                'AddressStatus': {'type': 'string'}
            }
        }
    },
    'Attributes': CUSTOM_PROPERTY_LIST,
    'CreatedDate': CREATED_DATE_FIELD,
    'CustomerKey': CUSTOMER_KEY_FIELD,
    'EmailAddress': {
        'type': ['null', 'string'],
        'description': ('Contains the email address for a subscriber. '
                        'Indicates the data extension field contains '
                        'email address data.'),
    },
    'EmailTypePreference': {
        'type': ['null', 'string'],
        'description': 'The format in which email should be sent'
    },
    'ID': ID_FIELD,
    'ListIDs': {
        'type': 'array',
        'description': 'Defines list IDs a subscriber resides on.',
        'items': {
            'type': 'string'
        }
    },
    'Locale': {
        'type': ['null', 'string'],
        'description': ('Contains the locale information for an Account. '
                        'If no location is set, Locale defaults to en-US '
                        '(English in United States).'),
    },
    'ModifiedDate': MODIFIED_DATE_FIELD,
    'ObjectID': OBJECT_ID_FIELD,
    'PartnerKey': {
        'type': ['null', 'string'],
        'description': ('Unique identifier provided by partner for an '
                        'object, accessible only via API.'),
    },
    'PartnerProperties': CUSTOM_PROPERTY_LIST,
    'PartnerType': {
        'type': ['null', 'string'],
        'description': 'Defines partner associated with a subscriber.'
    },
    'PrimaryEmailAddress': {
        'type': ['null', 'string'],
        'description': 'Indicates primary email address for a subscriber.'
    },
    'PrimarySMSAddress': {
        'type': ['null', 'string'],
        'description': ('Indicates primary SMS address for a subscriber. '
                        'Used to create and update SMS Address for a '
                        'given subscriber.'),
    },
    'PrimarySMSPublicationStatus': {
        'type': ['null', 'string'],
        'description': 'Indicates the subscriber\'s modality status.',
    },
    'Status': {
        'type': 'string',
        'description': 'Defines status of object. Status of an address.',
    },
    'SubscriberKey': SUBSCRIBER_KEY_FIELD,
    'SubscriberTypeDefinition': {
        'type': ['null', 'string'],
        'description': ('Specifies if a subscriber resides in an '
                        'integration, such as Salesforce or Microsoft '
                        'Dynamics CRM'),
    },
    'UnsubscribedDate': {
        'type': ['null', 'string'],
        'description': ('Represents date subscriber unsubscribed '
                        'from a list.'),
    }
})


class SubscriberDataAccessObject(DataAccessObject):

    SCHEMA = SCHEMA
    TABLE = 'subscriber'
    KEY_PROPERTIES = ['ID']

    def parse_object(self, obj):
        to_return = obj.copy()

        if 'ListIDs' in to_return:
            to_return['ListIDs'] = [_list.get('ObjectID')
                                    for _list in to_return.get('Lists', [])]

        if 'Lists' in to_return:
            del to_return['Lists']

        if to_return.get('Addresses') is None:
            to_return['Addresses'] = []

        if to_return.get('PartnerProperties') is None:
            to_return['PartnerProperties'] = []

        return super(SubscriberDataAccessObject, self).parse_object(obj)

    def sync_data(self):
        pass

    def pull_subscribers_batch(self, subscriber_keys):
        table = self.__class__.TABLE
        stream = request('Subscriber', FuelSDK.ET_Subscriber, self.auth_stub, {
            'Property': 'SubscriberKey',
            'SimpleOperator': 'IN',
            'Value': subscriber_keys
        })

        for subscriber in stream:
            subscriber = self.filter_keys_and_parse(subscriber)

            singer.write_records(table, [subscriber])



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.schemas import ID_FIELD, CUSTOM_PROPERTY_LIST, \
    CREATED_DATE_FIELD, MODIFIED_DATE_FIELD, CUSTOMER_KEY_FIELD, \
    OBJECT_ID_FIELD, with_properties
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table


LOGGER = singer.get_logger()


class ContentAreaDataAccessObject(DataAccessObject):

    SCHEMA = with_properties({
        'BackgroundColor': {
            'type': ['null', 'string'],
            'description': 'Indicates background color of content area',
        },
        'BorderColor': {
            'type': ['null', 'string'],
            'description': ('Indicates color of border surrounding '
                            'content area'),
        },
        'BorderWidth': {
            'type': ['null', 'integer'],
            'description': ('Indicates pixel width of border '
                            'surrounding content area'),
        },
        'CategoryID': {
            'type': ['null', 'integer'],
            'description': 'Specifies the identifier of the folder.',
        },
        'Cellpadding': {
            'type': ['null', 'integer'],
            'description': ('Indicates pixel value of padding '
                            'around content area'),
        },
        'Cellspacing': {
            'type': ['null', 'integer'],
            'description': ('Indicates pixel value of spacing '
                            'for content area'),
        },
        'Content': {
            'type': ['null', 'string'],
            'description': ('Identifies content contained in '
                            'a content area.'),
        },
        'CreatedDate': CREATED_DATE_FIELD,
        'CustomerKey': CUSTOMER_KEY_FIELD,
        'FontFamily': {
            'type': ['null', 'string'],
            'description': 'Indicates font family used in content area',
        },
        'HasFontSize': {
            'type': 'boolean',
            'description': ('Indicates whether the content area includes '
                            'a specified font size or not'),
        },
        'ID': ID_FIELD,
        'IsBlank': {
            'type': 'boolean',
            'description': ('Indicates if specified content area '
                            'contains no content.'),
        },
        'IsDynamicContent': {
            'type': 'boolean',
            'description': ('Indicates if specific content area '
                            'contains dynamic content.'),
        },
        'IsLocked': {
            'type': 'boolean',
            'description': ('Indicates if specific email content area '
                            'within an Enterprise or Enterprise 2.0 '
                            'account is locked and cannot be changed by '
                            'subaccounts.'),
        },
        'IsSurvey': {
            'type': 'boolean',
            'description': ('Indicates whether a specific content area '
                            'contains survey questions.'),
        },
        'Key': {
            'type': ['null', 'string'],
            'description': ('Specifies key associated with content area '
                            'in HTML body. Relates to the Email object '
                            'via a custom type.'),
        },
        'ModifiedDate': MODIFIED_DATE_FIELD,
        'Name': {
            'type': ['null', 'string'],
            'description': 'Name of the object or property.',
        },
        'ObjectID': OBJECT_ID_FIELD,
        'PartnerProperties': CUSTOM_PROPERTY_LIST,
        'Width': {
            'type': ['null', 'integer'],
            'description': 'Indicates pixel width of content area',
        },
    })

    TABLE = 'content_area'
    KEY_PROPERTIES = ['ID']

    def sync_data(self):
        table = self.__class__.TABLE
        selector = FuelSDK.ET_ContentArea

        search_filter = None
        retrieve_all_since = get_last_record_value_for_table(self.state, table)

        if retrieve_all_since is not None:
            search_filter = {
                'Property': 'ModifiedDate',
                'SimpleOperator': 'greaterThan',
                'Value': retrieve_all_since
            }

        stream = request('ContentAreaDataAccessObject',
                         selector,
                         self.auth_stub,
                         search_filter)

        for content_area in stream:
            content_area = self.filter_keys_and_parse(content_area)

            self.state = incorporate(self.state,
                                     table,
                                     'ModifiedDate',
                                     content_area.get('ModifiedDate'))

            singer.write_records(table, [content_area])

        save_state(self.state)



import unittest

from tap_exacttarget.util import partition_all


class TestPartitionAll(unittest.TestCase):

    def test__partition_all(self):
        self.assertEqual(
            list(partition_all([1, 2, 3, 4, 5, 6, 7], 3)),
            [[1, 2, 3], [4, 5, 6], [7]])



import datetime
from tap_exacttarget.filters import between


DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


def before_today(date_value):
    return (datetime.datetime.strptime(date_value, DATE_FORMAT).date() <=
            datetime.datetime.today().date())


def increment_date(date_value, unit=None):
    if unit is None:
        unit = {'days': 1}

    date_obj = datetime.datetime.strptime(date_value, DATE_FORMAT)

    incremented_date_obj = date_obj + datetime.timedelta(**unit)

    return datetime.datetime.strftime(incremented_date_obj, DATE_FORMAT)


def get_date_page(field, start, unit):
    return between(field, start, increment_date(start, unit))



import unittest

from tap_exacttarget.state import incorporate


class TestState(unittest.TestCase):

    def test_incorporate(self):
        self.assertEqual(
            incorporate({}, 'table', 'modifieddate', '2017-11-01'),
            {
                'bookmarks': {
                    'table': {
                        'last_record': '2017-11-01T00:00:00Z',
                        'field': 'modifieddate'
                    }
                }
            })

        self.assertEqual(
            incorporate({
                'bookmarks': {
                    'table': {
                        'last_record': '2017-01-01T00:00:00Z',
                        'field': 'modifieddate'
                    }
                }
            }, 'table', 'modifieddate', '2017-11-01'),
            {
                'bookmarks': {
                    'table': {
                        'last_record': '2017-11-01T00:00:00Z',
                        'field': 'modifieddate'
                    }
                }
            })

        self.assertEqual(
            incorporate({
                'bookmarks': {
                    'table': {
                        'last_record': '2017-11-01T00:00:00Z',
                        'field': 'modifieddate'
                    }
                }
            }, 'table', 'modifieddate', '2017-01-01'),
            {
                'bookmarks': {
                    'table': {
                        'last_record': '2017-11-01T00:00:00Z',
                        'field': 'modifieddate'
                    }
                }
            })



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.schemas import ID_FIELD, CUSTOM_PROPERTY_LIST, \
    CREATED_DATE_FIELD, CUSTOMER_KEY_FIELD, MODIFIED_DATE_FIELD, \
    DESCRIPTION_FIELD, OBJECT_ID_FIELD, with_properties
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table


LOGGER = singer.get_logger()


class FolderDataAccessObject(DataAccessObject):

    SCHEMA = with_properties({
        'AllowChildren': {
            'type': 'boolean',
            'description': ('Specifies whether a data folder can have '
                            'child data folders.'),
        },
        'ContentType': {
            'type': 'string',
            'description': ('Defines the type of content contained '
                            'within a folder.'),
        },
        'CreatedDate': CREATED_DATE_FIELD,
        'CustomerKey': CUSTOMER_KEY_FIELD,
        'Description': DESCRIPTION_FIELD,
        'ID': ID_FIELD,
        'ModifiedDate': MODIFIED_DATE_FIELD,
        'Name': {
            'type': ['null', 'string'],
            'description': 'Name of the object or property.',
        },
        'ObjectID': OBJECT_ID_FIELD,
        'ParentFolder': {
            'type': ['null', 'integer'],
            'description': ('Specifies the parent folder for a data '
                            'folder.'),
        },
        'PartnerProperties': CUSTOM_PROPERTY_LIST,
        'Type': {
            'type': ['null', 'string'],
            'description': ('Indicates type of specific list. Valid '
                            'values include Public, Private, Salesforce, '
                            'GlobalUnsubscribe, and Master.')
        }
    })

    TABLE = 'folder'
    KEY_PROPERTIES = ['ID']

    def parse_object(self, obj):
        to_return = obj.copy()

        to_return['ParentFolder'] = to_return.get('ParentFolder', {}).get('ID')

        return super(FolderDataAccessObject, self).parse_object(to_return)

    def sync_data(self):
        table = self.__class__.TABLE
        selector = FuelSDK.ET_Folder

        search_filter = None

        retrieve_all_since = get_last_record_value_for_table(self.state, table)

        if retrieve_all_since is not None:
            search_filter = {
                'Property': 'ModifiedDate',
                'SimpleOperator': 'greaterThan',
                'Value': retrieve_all_since
            }

        stream = request('Folder',
                         selector,
                         self.auth_stub,
                         search_filter)

        for folder in stream:
            folder = self.filter_keys_and_parse(folder)

            self.state = incorporate(self.state,
                                     table,
                                     'ModifiedDate',
                                     folder.get('ModifiedDate'))

            singer.write_records(table, [folder])

        save_state(self.state)



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.endpoints.subscribers import SubscriberDataAccessObject
from tap_exacttarget.pagination import get_date_page, before_today, \
    increment_date
from tap_exacttarget.schemas import ID_FIELD, CUSTOM_PROPERTY_LIST, \
    CREATED_DATE_FIELD, OBJECT_ID_FIELD, MODIFIED_DATE_FIELD, \
    SUBSCRIBER_KEY_FIELD, with_properties
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table
from tap_exacttarget.util import partition_all, sudsobj_to_dict


LOGGER = singer.get_logger()


def _get_subscriber_key(list_subscriber):
    return list_subscriber.SubscriberKey


def _get_list_subscriber_filter(_list, start, unit):
    return {
        'LogicalOperator': 'AND',
        'LeftOperand': {
            'Property': 'ListID',
            'SimpleOperator': 'equals',
            'Value': _list.get('ID'),
        },
        'RightOperand': get_date_page('ModifiedDate', start, unit)
    }


class ListSubscriberDataAccessObject(DataAccessObject):
    SCHEMA = with_properties({
        'ID': ID_FIELD,
        'CreatedDate': CREATED_DATE_FIELD,
        'ModifiedDate': MODIFIED_DATE_FIELD,
        'ObjectID': OBJECT_ID_FIELD,
        'PartnerProperties': CUSTOM_PROPERTY_LIST,
        'ListID': {
            'type': ['null', 'integer'],
            'description': ('Defines identification for a list the '
                            'subscriber resides on.'),
        },
        'Status': {
            'type': 'string',
            'description': ('Defines status of object. Status of '
                            'an address.'),
        },
        'SubscriberKey': SUBSCRIBER_KEY_FIELD,
    })

    TABLE = 'list_subscriber'
    KEY_PROPERTIES = ['SubscriberKey', 'ListID']

    def __init__(self, config, state, auth_stub, catalog):
        super(ListSubscriberDataAccessObject, self).__init__(
            config, state, auth_stub, catalog)

        self.replicate_subscriber = False
        self.subscriber_catalog = None

    def _get_all_subscribers_list(self):
        """
        Find the 'All Subscribers' list via the SOAP API, and return it.
        """
        result = request('List', FuelSDK.ET_List, self.auth_stub, {
            'Property': 'ListName',
            'SimpleOperator': 'equals',
            'Value': 'All Subscribers',
        })

        lists = list(result)

        if len(lists) != 1:
            msg = ('Found {} all subscriber lists, expected one!'
                   .format(len(lists)))
            raise RuntimeError(msg)

        return sudsobj_to_dict(lists[0])

    def sync_data(self):
        table = self.__class__.TABLE
        subscriber_dao = SubscriberDataAccessObject(
            self.config,
            self.state,
            self.auth_stub,
            self.subscriber_catalog)

        start = get_last_record_value_for_table(self.state, table)

        if start is None:
            start = self.config.get('start_date')

        pagination_unit = self.config.get(
            'pagination__list_subscriber_interval_unit', 'days')
        pagination_quantity = self.config.get(
            'pagination__list_subsctiber_interval_quantity', 1)

        unit = {pagination_unit: int(pagination_quantity)}

        end = increment_date(start, unit)

        all_subscribers_list = self._get_all_subscribers_list()

        while before_today(start):
            stream = request('ListSubscriber',
                             FuelSDK.ET_List_Subscriber,
                             self.auth_stub,
                             _get_list_subscriber_filter(
                                 all_subscribers_list,
                                 start, unit))

            batch_size = 100

            if self.replicate_subscriber:
                subscriber_dao.write_schema()

            for list_subscribers_batch in partition_all(list(stream),
                                                        batch_size):
                for list_subscriber in list_subscribers_batch:
                    list_subscriber = self.filter_keys_and_parse(
                        list_subscriber)

                    if list_subscriber.get('ModifiedDate'):
                        self.state = incorporate(
                            self.state,
                            table,
                            'ModifiedDate',
                            list_subscriber.get('ModifiedDate'))

                    singer.write_records(table, [list_subscriber])

                if self.replicate_subscriber:
                    subscriber_keys = list(map(
                        _get_subscriber_key, list_subscribers_batch))

                    subscriber_dao.pull_subscribers_batch(subscriber_keys)

            save_state(self.state)

            start = end
            end = increment_date(start, unit)



import FuelSDK
import singer

from tap_exacttarget.client import request
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.schemas import with_properties

LOGGER = singer.get_logger()


class CampaignDataAccessObject(DataAccessObject):

    SCHEMA = with_properties({
        'id': {
            'type': 'string',
        },
        'createdDate': {
            'type': 'string',
        },
        'modifiedDate': {
            'type': 'string',
        },
        'name': {
            'type': 'string',
        },
        'description': {
            'type': 'string',
        },
        'campaignCode': {
            'type': 'string',
        },
        'color': {
            'type': 'string',
        }
    })

    TABLE = 'campaign'
    KEY_PROPERTIES = ['id']

    def sync_data(self):
        cursor = request(
            'Campaign',
            FuelSDK.ET_Campaign,
            self.auth_stub)

        for campaign in cursor:
            campaign = self.filter_keys_and_parse(campaign)

            singer.write_records(self.__class__.TABLE, [campaign])



import FuelSDK
import singer


LOGGER = singer.get_logger()


def _get_response_items(response):
    items = response.results

    if 'count' in response.results:
        LOGGER.info('Got {} results.'.format(response.results.get('count')))
        items = response.results.get('items')

    return items


__all__ = ['get_auth_stub', 'request', 'request_from_cursor']


# PUBLIC FUNCTIONS

def get_auth_stub(config):
    """
    Given a config dict in the format:

        {'clientid': ... your ET client ID ...,
         'clientsecret': ... your ET client secret ...}

    ... return an auth stub to be used when making requests.
    """
    LOGGER.info("Generating auth stub...")

    auth_stub = FuelSDK.ET_Client(
        params={
            'clientid': config['client_id'],
            'clientsecret': config['client_secret']
        })
    auth_stub.soap_client.set_options(
        timeout=int(config.get('request_timeout', 900)))

    LOGGER.info("Success.")

    return auth_stub


def request(name, selector, auth_stub, search_filter=None, props=None):
    """
    Given an object name (`name`), used for logging purposes only,
      a `selector`, for example FuelSDK.ET_ClickEvent,
      an `auth_stub`, generated by `get_auth_stub`,
      an optional `search_filter`,
      and an optional set of `props` (properties), which specifies the fields
        to be returned from this object,

    ... request data from the ExactTarget API using FuelSDK. This function
    returns a generator that will yield all the records returned by the
    request.

    Example `search_filter`:

        {'Property': 'CustomerKey',
         'SimpleOperator': 'equals',
         'Value': 'abcdef'}

    For more on search filters, see:
      https://developer.salesforce.com/docs/atlas.en-us.noversion.mc-apis.meta/mc-apis/using_complex_filter_parts.htm
    """
    cursor = selector()
    cursor.auth_stub = auth_stub

    if props is not None:
        cursor.props = props

    if search_filter is not None:
        cursor.search_filter = search_filter

        LOGGER.info(
            "Making RETRIEVE call to '{}' endpoint with filters '{}'."
            .format(name, search_filter))

    else:
        LOGGER.info(
            "Making RETRIEVE call to '{}' endpoint with no filters."
            .format(name))

    return request_from_cursor(name, cursor)


def request_from_cursor(name, cursor):
    """
    Given an object name (`name`), used for logging purposes only, and a
    `cursor` provided by FuelSDK, return a generator that yields all the
    items in that cursor.

    Primarily used internally by `request`, but can be used if cursors have
    to be customized. See tap_exacttarget.endpoints.data_extensions for
    an example.
    """
    response = cursor.get()

    if not response.status:
        raise RuntimeError("Request failed with '{}'"
                           .format(response.message))

    for item in _get_response_items(response):
        yield item

    while response.more_results:
        LOGGER.info("Getting more results from '{}' endpoint".format(name))

        response = cursor.getMoreResults()

        if not response.status:
            raise RuntimeError("Request failed with '{}'"
                               .format(response.message))

        for item in _get_response_items(response):
            yield item

    LOGGER.info("Done retrieving results from '{}' endpoint".format(name))



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='tap-exacttarget',
    version='0.0.7',
    description='Singer.io tap for extracting data from the ExactTarget API',
    author='Fishtown Analytics',
    url='http://fishtownanalytics.com',
    classifiers=['Programming Language :: Python :: 3 :: Only'],
    py_modules=['tap_exacttarget'],
    install_requires=[
        'funcy==1.9.1',
        'singer-python>=3.5.0',
        'python-dateutil==2.6.0',
        'voluptuous==0.10.5',
        'pyjwt>=0.1.9',
        'requests>=2.2.1',
        'suds-jurko>=0.6'
    ],
    entry_points='''
    [console_scripts]
    tap-exacttarget=tap_exacttarget:main
    ''',
    packages=find_packages()
)



import FuelSDK
import singer

from funcy import set_in, update_in, merge

from tap_exacttarget.client import request, request_from_cursor
from tap_exacttarget.dao import DataAccessObject
from tap_exacttarget.pagination import get_date_page, before_today, \
    increment_date
from tap_exacttarget.state import incorporate, save_state, \
    get_last_record_value_for_table
from tap_exacttarget.util import sudsobj_to_dict

LOGGER = singer.get_logger()  # noqa


def _merge_in(collection, path, new_item):
    return update_in(collection, path, lambda x: merge(x, [new_item]))


def _convert_extension_datatype(datatype):
    if datatype in ['Boolean']:
        return 'boolean'
    elif datatype in ['Decimal', 'Number']:
        return 'number'

    return 'string'


def _convert_data_extension_to_catalog(extension):
    return {
        field.get('Name'): {
            'type': _convert_extension_datatype(field.get('ValueType')),
            'description': field.get('Description'),
            'inclusion': 'available',
        }
        for field in extension.get('Fields')
    }


def _get_tap_stream_id(extension):
    extension_name = extension.CustomerKey
    return 'data_extension.{}'.format(extension_name)


def _get_extension_name_from_tap_stream_id(tap_stream_id):
    return tap_stream_id.split('.')[1]


class DataExtensionDataAccessObject(DataAccessObject):

    @classmethod
    def matches_catalog(cls, catalog):
        return 'data_extension.' in catalog.get('stream')

    def _get_extensions(self):
        result = request(
            'DataExtension',
            FuelSDK.ET_DataExtension,
            self.auth_stub,
            props=['CustomerKey', 'Name'])

        to_return = {}

        for extension in result:
            extension_name = str(extension.Name)
            customer_key = str(extension.CustomerKey)

            to_return[customer_key] = {
                'tap_stream_id': 'data_extension.{}'.format(customer_key),
                'stream': 'data_extension.{}'.format(extension_name),
                'key_properties': ['_CustomObjectKey'],
                'schema': {
                    'type': 'object',
                    'inclusion': 'available',
                    'selected-by-default': False,
                    'properties': {
                        '_CustomObjectKey': {
                            'type': 'string',
                            'description': ('Hidden auto-incrementing primary '
                                            'key for data extension rows.'),
                        },
                        'CategoryID': {
                            'type': 'integer',
                            'description': ('Specifies the identifier of the '
                                            'folder. (Taken from the parent '
                                            'data extension.)')
                        }
                    }
                },
            }

        return to_return

    def _get_fields(self, extensions):
        to_return = extensions.copy()

        result = request(
            'DataExtensionField',
            FuelSDK.ET_DataExtension_Column,
            self.auth_stub)

        for field in result:
            extension_id = field.DataExtension.CustomerKey
            field = sudsobj_to_dict(field)
            field_name = field['Name']

            if field.get('IsPrimaryKey'):
                to_return = _merge_in(
                    to_return,
                    [extension_id, 'key_properties'],
                    field_name)

            field_schema = {
                'type': [
                    'null',
                    _convert_extension_datatype(str(field.get('FieldType')))
                ],
                'description': str(field.get('Description')),
            }

            to_return = set_in(
                to_return,
                [extension_id, 'schema', 'properties', field_name],
                field_schema)

        return to_return

    def generate_catalog(self):
        # get all the data extensions by requesting all the fields
        extensions_catalog = self._get_extensions()

        extensions_catalog_with_fields = self._get_fields(extensions_catalog)

        return extensions_catalog_with_fields.values()

    def parse_object(self, obj):
        properties = obj.get('Properties', {}).get('Property', {})
        to_return = {}

        for prop in properties:
            to_return[prop['Name']] = prop['Value']

        return to_return

    def _replicate(self, customer_key, keys,
                   parent_category_id, table,
                   partial=False, start=None,
                   end=None, unit=None, replication_key=None):
        if partial:
            LOGGER.info("Fetching {} from {} to {}"
                        .format(table, start, end))

        cursor = FuelSDK.ET_DataExtension_Row()
        cursor.auth_stub = self.auth_stub
        cursor.CustomerKey = customer_key
        cursor.props = keys

        if partial:
            cursor.search_filter = get_date_page(replication_key,
                                                 start,
                                                 unit)

        result = request_from_cursor('DataExtensionObject', cursor)

        for row in result:
            row = self.filter_keys_and_parse(row)
            row['CategoryID'] = parent_category_id

            self.state = incorporate(self.state,
                                     table,
                                     replication_key,
                                     row.get(replication_key))

            singer.write_records(table, [row])

        if partial:
            self.state = incorporate(self.state,
                                     table,
                                     replication_key,
                                     start)

            save_state(self.state)

    def sync_data(self):
        tap_stream_id = self.catalog.get('tap_stream_id')
        table = self.catalog.get('stream')
        (_, customer_key) = tap_stream_id.split('.', 1)

        keys = self.get_catalog_keys()
        keys.remove('CategoryID')

        replication_key = None

        start = get_last_record_value_for_table(self.state, table)

        if start is None:
            start = self.config.get('start_date')

        for key in ['ModifiedDate', 'JoinDate']:
            if key in keys:
                replication_key = key

        pagination_unit = self.config.get(
            'pagination__data_extension_interval_unit', 'days')
        pagination_quantity = self.config.get(
            'pagination__data_extension_interval_quantity', 7)

        unit = {pagination_unit: int(pagination_quantity)}

        end = increment_date(start, unit)

        parent_result = None
        parent_extension = None
        parent_result = request(
            'DataExtension',
            FuelSDK.ET_DataExtension,
            self.auth_stub,
            search_filter={
                'Property': 'CustomerKey',
                'SimpleOperator': 'equals',
                'Value': customer_key,
            },
            props=['CustomerKey', 'CategoryID'])

        parent_extension = next(parent_result)
        parent_category_id = parent_extension.CategoryID

        while before_today(start) or replication_key is None:
            self._replicate(
                customer_key,
                keys,
                parent_category_id,
                table,
                partial=(replication_key is not None),
                start=start,
                end=end,
                replication_key=replication_key)

            if replication_key is None:
                return

            self.state = incorporate(self.state,
                                     table,
                                     replication_key,
                                     start)

            save_state(self.state)

            start = end
            end = increment_date(start, unit)
